/* Implementation of XtGState class
 *
 * Copyright (C)  1995  The Board of Trustees of  
 * The Leland Stanford Junior University.  All Rights Reserved.
 *
 * Authors: Paul Kunz
 *
 * This file is part of an Objective-C class library for a X-Window System
 *
 * XtGState.m,v 1.8 1995/04/27 21:54:49 pfkeb Exp
 */

#include "XtGState.h"

#include "appkit/Application.h"
#include "appkit/View.h"
#include "appkit/Window.h"
#include "appkit/xttools.h"
#include <objc/List.h>

/* Private functions we use from Window and View */
@interface Window(WindowPrivate)
- (Pixmap)_backingPixmap;
@end
@interface View(WidgetSet)
- _keepGState:(XtGState *)gstate;
@end

static List     *gstates = nil;

@implementation XtGState

void
GSNeedsDisplay(XWContext context)
{
    XtGState	*gstate = (XtGState *)context->gstate;
    
    if ( !gstate ) 
    	return;
    [gstate setNeedsDisplay:YES];
}

void
GSPing(NXRect *frame)
{
    XWContext   context = (XWContext)[NXApp context];
    XtGState	*gstate;

    gstate = (XtGState *)context->gstate;    
    if (!gstate) 
    	return;
    [gstate flushRect:frame];
}

+ initialize
{
  /* gstate index of 0 is indicator of no gstate, thus we put
   * dummy at index 0
   */
    gstates = [[List alloc] initCount:1];
    [gstates addObject:[[Object alloc] init]];
    return self;
}

+ alloc
{
    return [self allocFromZone:(NSZone *)0];
}

+ allocFromZone:(NSZone *)zone
{
    XtGState *newgs = [super allocFromZone:zone];

    [gstates addObject:newgs];
    return newgs;
}

+ removeFromList:(XtGState *)aGState
{
    unsigned int i = [gstates indexOf:aGState];

    if ( i == NX_NOT_IN_LIST ) {
	fprintf(stderr, 
		"XtGState error: attempt to free gstate not in list\n");
        return self;
    }

    if ( i == ([gstates count] - 1) ) {
        [gstates removeLastObject];
    } else {
        Object *dummy = [[Object alloc] init];
        [gstates replaceObjectAt:i with:dummy];
    }
    return self;
}

+ (int)objectToNum:(XtGState *)aGState
{
    unsigned int i = [gstates indexOf:aGState];

    if ( i == NX_NOT_IN_LIST ) {
        fprintf(stderr, 
             "XtGState error: attempt to find non-existent gstate object\n");
    }
    return i;
}

+ numToObject:(int)aNum
{
    XtGState *aGState = [gstates objectAt:aNum];

    if ( [aGState isKindOf:[XtGState class]] ) {
	return aGState;
    } else {
	fprintf(stderr, 
	        "XtGState error: can't convert gstate num to object\n");
	aGState = nil;
    }
    return aGState;
}

- initFromSuper:(XtGState *)superGState withView:(View *)aView
{
    cgt_t super_cgt;
    view = aView;
    widget = [aView _widget];

    if (superGState)
     	super_cgt = [superGState currentCGT];
    else
    	super_cgt = composeCGT(NULL, 0, NULL);
	
    /* Set the Drawable we are supposed to draw too.  If the window we
       are in is using backing store, we need to draw to the backing pixmap
       and then transfer the data later on.
    */
    draw = [[aView window] _backingPixmap];
    if (draw == 0) {
	if (widget)
	    draw = XtWindow(widget);
	[aView getBounds:&cliprect];
	cgt = composeCGT(NULL, 0, NULL);
    } else {
    	/* Get our bounds in window coordinates */
	NXPoint max;
	[aView getFrame:&cliprect];
	cgt = concatCGT(super_cgt, composeCGT(NULL, 0, &cliprect.origin));
	max.x = NX_MAXX(&cliprect);
	max.y = NX_MAXY(&cliprect);
	cliprect.origin = applyCGT(cliprect.origin, super_cgt);
	max = applyCGT(max, super_cgt);
	NX_WIDTH(&cliprect) = max.x - NX_X(&cliprect);
	NX_HEIGHT(&cliprect) = max.y - NX_Y(&cliprect);
	if (superGState) {
	    NXRect super_clip;
	    [superGState getClipRect:&super_clip];
	    NXIntersectionRect(&super_clip, &cliprect);
	}
    }
    update = NO;

    return self;
}

- free
{
    XWContext context = (XWContext)[NXApp context];

    if ( sharedGC == NO && xgcntxt ) {
	XFreeGC(context->display, xgcntxt);
    }
    [XtGState removeFromList:self];
    return [super free];
}

- setGraphicContext:(GC)context shared:(BOOL)flag
{
  /* If the GC is given to receiving object from the GC of another, 
   * then should not be freed when object is freed.
   */
    sharedGC = flag;
    xgcntxt = context;
    return self;
}

- (GC)graphicContext
{
    return xgcntxt;
}

- (BOOL)hasGraphicContext
{
    return xgcntxt ? YES : NO;
}

- copyGraphicContext: (XWContext) context
{
    GC                  source;
    unsigned long	mask;

    source = [(XtGState *)context->gstate graphicContext];
    mask = 0xffffffff; /* Copy everything (Hopefully) */
    if ( XtIsRealized(widget) ) {
	xgcntxt = XCreateGC(context->display, 
			    XtWindow(widget), 0, NULL);
    } else {
	xgcntxt = XCreateGC(context->display, 
			    DefaultRootWindow(context->display), 0, NULL );
    }
    XCopyGC(context->display, source, mask, xgcntxt); 
    return self;
}

// Create a default graphics context.
- createGraphicContext:(XWContext) context
{
    XGCValues gcv;
    Pixel		black, white;

    if (!widget) {
	fprintf(stderr, 
            "Error (View): Attempt to create a GC on an unrealized View\n");
	return nil;
    }

    gcv.function = GXcopy;
//    gcv.foreground = xtGrayToPixel(context, 0.0);	/* NX_BLACK */
//    gcv.background = xtGrayToPixel(context, 0.6666);	/* NX_LTGRAY */
    black = XBlackPixel(context->display, XDefaultScreen(context->display));
    white = XWhitePixel(context->display, XDefaultScreen(context->display));
    gcv.background = white;
    gcv.foreground = black;
    gcv.plane_mask = AllPlanes;
    gcv.fill_style = FillSolid;
    if ( XtIsRealized(widget) ) {
	xgcntxt = 
	    XCreateGC(context->display, XtWindow(widget),
	      GCFunction | GCForeground | GCBackground | GCPlaneMask, &gcv);
    } else {
	xgcntxt = 
	    XCreateGC(context->display, DefaultRootWindow(context->display),
         GCFillStyle | GCFunction | GCForeground | GCBackground | GCPlaneMask, 
		&gcv);
    }
    return self;
}

- setGray:(float)gray
{
    XWContext	context = (XWContext)[NXApp context];
    XGCValues	gcv;
    Pixel	value;
    
    value = xtGrayToPixel(context, gray);
    gcv.foreground = value;
    XChangeGC(context->display, xgcntxt, GCForeground, &gcv);
    return self;
}

- translate:(const NXPoint *)point
{
    cgt = concatCGT(cgt, composeCGT(NULL, 0, point));
    return self;
}

- rotate:(float)degrees
{
    cgt = concatCGT(cgt, composeCGT(NULL, degrees, NULL));
    return self;
}

- scale:(NXSize)scale
{
    cgt = concatCGT(cgt, composeCGT(&scale, 0, NULL));
    return self;
}

- concatCGT:(cgt_t)acgt;
{
    cgt = concatCGT(cgt, acgt);
    return self;
}

- (NXPoint)offset
{
    return offsetFromCGT(cgt);
}

- (NXSize)scale
{
    return scaleFromCGT(cgt);
}

- (float)rotation
{
    return rotationFromCGT(cgt);
}

- (cgt_t)currentCGT
{
    return cgt;
}

- getClipRect:(NXRect *)aRect;
{
    if ( aRect ) {
    	*aRect = cliprect;
    } else {
	fprintf(stderr, "Error (XtGState): -getClipRect given null rect\n");
    }
    return self;
}

- setNeedsDisplay:(BOOL)flag
{
    update = flag;
    return self;
}

- (BOOL)needsDisplay
{
   return update;
}

- (BOOL) didDraw
{
    return didDraw;
}

- (BOOL)hasWidget
{
    return (widget ? YES : NO);
}

- (Drawable)drawable
{
    return draw;
}

- (BOOL) hasDrawable
{
    return (draw ? YES : NO);
}

- copyDrawable:(Drawable)source fromRect:(NXRect *)aRect
{
    NXPoint	point;

    point.x = 0; point.y = 0;
    [self copyDrawable:source fromRect:aRect toPoint:&point];
    return self;
};
    
- setDidUpdate
{
    update = YES;
    [view _keepGState:self];
    return self;
}

- copyDrawable:(Drawable)source fromRect:(NXRect *)aRect 
			         toPoint:(NXPoint *)aPoint
{
    XWContext	context = (XWContext)[NXApp context];
    
    [self convertPointToDrawable:aPoint];
    aPoint->y -= NX_HEIGHT(aRect);
    XCopyArea(context->display, source, draw, xgcntxt,
		aRect->origin.x,   aRect->origin.y,
		aRect->size.width, aRect->size.height,
		aPoint->x, aPoint->y);
    [self setDidUpdate];
    return self;
}

- copyImage:(XImage *)image ofSize:(NXSize *)size
{
    NXPoint	point;
    XWContext	context = (XWContext)[NXApp context];

    point.x = 0; point.y = 0;
    [self convertPointToDrawable:&point];
    point.y -= size->height;
    XPutImage(context->display, draw, xgcntxt,
			image, 	0, 0,
			point.x, point.y,
			size->width, size->height);
    [self setDidUpdate];
    return self;
}

- convertRectToDrawable:(NXRect *)aRect
{
    NXRect bounds;
    NXPoint max;
    
    max.x = NX_MAXX(aRect);
    max.y = NX_MAXY(aRect);
    aRect->origin = applyCGT(aRect->origin, cgt);
    max = applyCGT(max, cgt);
    NX_WIDTH(aRect) = max.x - NX_X(aRect);
    NX_HEIGHT(aRect) = max.y - NX_Y(aRect);
    
    /* Flip the coordinate system for X-windows */
    if (draw == XtWindow(widget))
	[view getBounds:&bounds];
    else
	[[view window] getFrame:&bounds];
    NX_Y(aRect) = NX_HEIGHT(&bounds) - NX_MAXY(aRect);

    return self;
}

- convertPointToDrawable:(NXPoint *)aPoint
{
    NXRect bounds;
    
    *aPoint = applyCGT(*aPoint, cgt);
    
    /* Flip the coordinate system for X-windows */
    if (draw == XtWindow(widget))
	[view getBounds:&bounds];
    else
	[[view window] getFrame:&bounds];
    aPoint->y = NX_HEIGHT(&bounds) - aPoint->y;

    return self;
}

- rectFill:(const NXRect *)aRect
{
    XWContext	context = (XWContext)[NXApp context];
    NXRect	bounds;

    bounds = *aRect;
    [self convertRectToDrawable:&bounds];
    XFillRectangle(context->display, draw, xgcntxt,
                   NX_X(&bounds), NX_Y(&bounds),
		   NX_WIDTH(&bounds), NX_HEIGHT(&bounds));
    [self setDidUpdate];
    return self;
}

- (void)flushRect:(NXRect *)frame
{
    XWContext   context;
    NXRect	bounds;

  /* Are we doing backing? */
    if (xgcntxt == 0)
	return;
    if (widget == 0)
	return;
    if (!XtIsRealized(widget))
	return;
    if (draw == XtWindow(widget))
	return;

    context = (XWContext)[NXApp context];
    bounds = *frame;
    [self convertRectToDrawable:&bounds];
    XCopyArea(context->display, draw, XtWindow(widget), xgcntxt,
		NX_X(&bounds), NX_Y(&bounds),
		NX_WIDTH(&bounds), NX_HEIGHT(&bounds),
		NX_X(&bounds), NX_Y(&bounds));
    update = NO;
    didDraw = YES;
    return;
}

- (void)flushRectIfNeeded:(NXRect *)frame
{
    if ( update ) {
	return [self flushRect:frame];
    } 
    return;
}

- (BOOL) hasSameWidgetAs:(View *)aView
{
    return (widget == [aView _widget]);
}

@end
