/*
    cgt - Functions to deal with CGT matrices
    
    Copyright (c) 1993, Adam Fedor.
*/

#ifndef _CGT_INCLUDE_
#define _CGT_INCLUDE_

#include "appkit/graphics.h"

/* 
    Homogenous coordinate geometric transform matrix
	[ a  b  0 ]
    T = [ c  d  0 ]
	[ tx ty 1 ]
    Transform a point by pre-multiplying the point by the matrix, i.e.
    	Pnew = transpose(P) * T
*/
typedef struct {
    double  a, b, c, d, tx, ty;
} cgt_t;

/* Useful functions to deal with CGT matrices. Rotation is specified
   in degrees.
	concatCGT  - concatenates two CGTs, acgt and bcgt, and
	returns the new matrix. (cgt = acgt * bcgt)

	composeCGT - Compose a CGT matrix from the 
	specified scale, translation, and rotation. The transformation is
	applied in the order that the arguments are given (i.e. scale, 
	then rotate, then translate). Arguments are given as pointers to make
	it easy to generate an identity transform matrix - use
	composeCGT(NULL, 0, NULL) to generate a matrix with no translation
	or rotation and a scaling of one.
	
	decomposeCGT - Return the transfomation information that was used to
	construct the CGT (via a call to composeCGT).
	
	applyCGT - Transforms the point by the given CGT. Returns the new 
	point.
*/
extern cgt_t	concatCGT(cgt_t acgt, cgt_t bcgt);
extern cgt_t	composeCGT(const NXSize *scale, double rotation, 
		const NXPoint *trans);
extern void	decomposeCGT(cgt_t cgt, NXSize *scale, double *rotation,
		NXPoint *trans);
extern cgt_t	invertCGT(cgt_t cgt);

extern NXPoint  applyCGT(NXPoint point, cgt_t cgt);
extern double	rotationFromCGT(cgt_t cgt);
extern NXSize	scaleFromCGT(cgt_t cgt);
extern NXPoint	offsetFromCGT(cgt_t cgt);

#endif
